package com.jeremyparsons.imaging;

import java.awt.BorderLayout;
import java.awt.CardLayout;
import java.awt.Container;
import java.awt.Cursor;
import java.awt.FlowLayout;
import java.awt.GridLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.awt.event.WindowListener;
import java.io.File;

import javax.swing.ButtonGroup;
import javax.swing.JButton;
import javax.swing.JFileChooser;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JRadioButton;
import javax.swing.JScrollPane;
import javax.swing.JSeparator;
import javax.swing.JTextField;
import javax.swing.SwingConstants;
import javax.swing.SwingUtilities;
import javax.swing.WindowConstants;

/**
Provide a graphical user interface (GUI) around JPEGDirCopy class
to allow setting of all parameters without a terminal and command line.

If working from a double-clicked jar file, then may need to restart JVM with more memory.
Anyone smart enough to know how to start Java from the command line, should know about heap size options.

This class has a main method to allow it to be called from
within a jar as the class to be started.

@see emb.ebi.jparsons.JPEGDirCopy;
@see com.sun.image.codec.jpeg;

Copyright Jeremy D. Parsons 2002,2003,2013,2015
@author Jeremy D. Parsons jparsons@jeremyparsons.com 
@version 1.3 Date: 20150110

*/

public class JPEGCopierGUI extends JFrame implements CounterListener, Runnable{
	private static final long serialVersionUID = 828463359312837520L;
	// Standard defaults 
	private int desiredWidth = CopyModel.DEF_OUT_WIDTH;
	private int desiredHeight = CopyModel.DEF_OUT_HEIGHT;
	private float desiredQuality = CopyModel.DEF_QUALITY;
	private boolean desiredBaseline = CopyModel.DEF_BASELINE_ENCODING;
	private boolean desiredUltraQual = CopyModel.DEF_ULTRA_QUAL;

	public static String CHANGE_SRC_BUTTON_STR = "Choose folder to be copied";
	public static String CHANGE_DEST_BUTTON_STR = "Type new output folder name";
	public static String CHANGE_QUALITY_BUTTON_STR = "Change output picture quality";
	public static String CHANGE_WIDTH_BUTTON_STR = "Change maximum picture width";
	public static String CHANGE_HEIGHT_BUTTON_STR = "Change maximum picture height";
	public static String DEGRAIN_TOGGLE_LABEL_STR = "Perform degraining ? (slow & subtle)";
	public static String DEGRAIN_TRUE_LABEL_STR = "Yes";
	public static String DEGRAIN_FALSE_LABEL_STR = "No";
	public static String ULTRA_QUAL_TOGGLE_LABEL_STR = "Highest quality ? (slow)";
	public static String ULTRA_QUAL_TRUE_LABEL_STR = "Yes";
	public static String ULTRA_QUAL_FALSE_LABEL_STR = "No";
	public static String EVERYTHING_SET_QUESTION_STR = "Configure everything above, then press start";
	public static String EVERYTHING_SET_BUTTON_STR = "Start Copying";
	public static String COUNTER_TITLE_STR = "Count of copied images= ";
	public static String UNSET_DIR_STR = "Folder not set yet";

	private String defaultDirName = System.getProperty("user.dir");
	private String lastInputDirName = defaultDirName;
	private String lastOutputDirName = defaultDirName;
	
	// Set up two panels to be alternately displayed within a single panel.
	private static final String START_VIEW = "Start";
	private static final String RUNNING_VIEW = "Running";

	JPanel startButtonPanel = new JPanel(new FlowLayout(FlowLayout.LEFT));
	JPanel runningPanel = new JPanel(new FlowLayout(FlowLayout.LEFT));

	private CardLayout myLayout = new CardLayout();
	private JPanel myCardPanel = new JPanel(myLayout);

	final JLabel inputDirField = new JLabel(UNSET_DIR_STR);
	final JLabel outputDirField = new JLabel(UNSET_DIR_STR);
	final JTextField qualityField = new JTextField(Float.toString(desiredQuality), 5);
	final JTextField widthField = new JTextField(Integer.toString(desiredWidth), 5);
	final JTextField heightField = new JTextField(Integer.toString(desiredHeight), 5);
	final ButtonGroup degrainToggle = new ButtonGroup();
	final ButtonGroup ultraQualToggle = new ButtonGroup();

	JLabel squashedCounterLabel = null;

	File inputFolder;
	File outputFolder;

	Thread runningThread = null;

	public JPEGCopierGUI (){
		super("JPEG Directory Copier Controls");
		//super.setDefaultCloseOperation(WindowConstants.EXIT_ON_CLOSE);
		WindowListener winListener = new WindowAdapter(){
			public void windowClosing(WindowEvent e) {System.exit(0);}
		};
		this.addWindowListener(winListener);

		JPanel selectionPanel = new JPanel();
		selectionPanel.setLayout(new GridLayout(7,1));

		Container content = this.getContentPane();
		JScrollPane scrollPane = new JScrollPane(selectionPanel);
		content.add(scrollPane);

		// Create the panel to display and select the input folder of jpeg image files
		JPanel inputFolderPanel = new JPanel(new FlowLayout(FlowLayout.LEFT));
		final JButton changeInputButton = new JButton(CHANGE_SRC_BUTTON_STR);
		changeInputButton.addActionListener(new ActionListener(){
			public void actionPerformed(ActionEvent actEv){
				changeInputRequest();
				}
			});

		inputFolderPanel.add(changeInputButton);
		inputFolderPanel.add(inputDirField);
		selectionPanel.add(inputFolderPanel);

		// Create the panel to display and select the output folder of jpeg image files
		JPanel outputFolderPanel = new JPanel(new FlowLayout(FlowLayout.LEFT));
		final JButton changeOutputButton = new JButton(CHANGE_DEST_BUTTON_STR);
		changeOutputButton.addActionListener(new ActionListener(){
			public void actionPerformed(ActionEvent actEv){
				changeOutputRequest();
				}
			});

		outputFolderPanel.add(changeOutputButton);
		outputFolderPanel.add(outputDirField);
		selectionPanel.add(outputFolderPanel);
		
		// Create the panel to display and select the quality value
		JPanel qualityPanel = new JPanel(new FlowLayout(FlowLayout.LEFT));
		final JButton changeQualityButton = new JButton(CHANGE_QUALITY_BUTTON_STR);
		changeOutputButton.addActionListener(new ActionListener(){
			public void actionPerformed(ActionEvent actEv){
				changeQualityRequest();
				}
			});

		qualityField.addActionListener(new ActionListener(){
			public void actionPerformed(ActionEvent actEv){
				changeQualityRequest();
				}
			});

		qualityPanel.add(changeQualityButton);
		qualityField.setHorizontalAlignment(SwingConstants.RIGHT);
		qualityPanel.add(qualityField);
		selectionPanel.add(qualityPanel);
		
		// Create the panel to display and select the maximum image width
		JPanel widthPanel = new JPanel(new FlowLayout(FlowLayout.LEFT));
		final JButton changeWidthButton = new JButton(CHANGE_WIDTH_BUTTON_STR);
		changeWidthButton.addActionListener(new ActionListener(){
			public void actionPerformed(ActionEvent actEv){
				changeWidthRequest();
				}
			});

		widthField.addActionListener(new ActionListener(){
			public void actionPerformed(ActionEvent actEv){
				changeWidthRequest();
				}
			});
		widthPanel.add(changeWidthButton);
		widthField.setHorizontalAlignment(SwingConstants.RIGHT);
		widthPanel.add(widthField);
		selectionPanel.add(widthPanel);
		
		// Create the panel to display and select the maximum image height
		JPanel heightPanel = new JPanel(new FlowLayout(FlowLayout.LEFT));
		final JButton changeHeightButton = new JButton(CHANGE_HEIGHT_BUTTON_STR);
		changeHeightButton.addActionListener(new ActionListener(){
			public void actionPerformed(ActionEvent actEv){
				changeHeightRequest();
				}
			});

		heightField.addActionListener(new ActionListener(){
			public void actionPerformed(ActionEvent actEv){
				changeHeightRequest();
				}
			});
		heightPanel.add(changeHeightButton);
		heightField.setHorizontalAlignment(SwingConstants.RIGHT);
		heightPanel.add(heightField);
		selectionPanel.add(heightPanel);
		
		JPanel degrainPanel = new JPanel(new FlowLayout(FlowLayout.LEFT));
		JRadioButton tempGrainRadioButton;
		degrainPanel.add(new JLabel(DEGRAIN_TOGGLE_LABEL_STR));
		degrainPanel.add(tempGrainRadioButton = new JRadioButton(DEGRAIN_TRUE_LABEL_STR));
		tempGrainRadioButton.setActionCommand("true");
		degrainToggle.add(tempGrainRadioButton);
		tempGrainRadioButton.setSelected(CopyModel.DEF_DEGRAIN);
		degrainPanel.add(tempGrainRadioButton = new JRadioButton(DEGRAIN_FALSE_LABEL_STR));
		tempGrainRadioButton.setActionCommand("false");
		degrainToggle.add(tempGrainRadioButton);
		tempGrainRadioButton.setSelected(!CopyModel.DEF_DEGRAIN);
		//degrainToggle.setSelected(someButtonModel, JPEGDirCopy.DEF_DEGRAIN);
		selectionPanel.add(degrainPanel);
		
		JPanel ultraQualityPanel = new JPanel(new FlowLayout(FlowLayout.LEFT));
		JRadioButton ultraQualRadioButton;
		ultraQualityPanel.add(new JLabel(ULTRA_QUAL_TRUE_LABEL_STR));
		ultraQualityPanel.add(ultraQualRadioButton = new JRadioButton(ULTRA_QUAL_TRUE_LABEL_STR));
		ultraQualRadioButton.setActionCommand("true");
		ultraQualToggle.add(ultraQualRadioButton);
		ultraQualityPanel.add(ultraQualRadioButton = new JRadioButton(ULTRA_QUAL_FALSE_LABEL_STR));
		ultraQualRadioButton.setActionCommand("false");
		ultraQualToggle.add(ultraQualRadioButton);
		ultraQualRadioButton.setSelected(CopyModel.DEF_ULTRA_QUAL);
	
		// Once all the selection panels are built, add the start button
		JPanel startButtonSeparatorPanel = new JPanel(new BorderLayout());
		startButtonSeparatorPanel.add(new JSeparator(SwingConstants.HORIZONTAL), BorderLayout.NORTH);
		final JButton startButton = new JButton(EVERYTHING_SET_BUTTON_STR);
		startButton.addActionListener(new ActionListener(){
			public void actionPerformed(ActionEvent actEv){
				// Check all values are drawn from text fields
				changeWidthRequest();
				changeHeightRequest();
				changeQualityRequest();

				// Start the processing thread
				runningThread = new Thread(JPEGCopierGUI.this);
				runningThread.start();
				}
			});

		final JLabel StartQuestionField = new JLabel(EVERYTHING_SET_QUESTION_STR);
		startButtonPanel.add(StartQuestionField);
		startButtonPanel.add(startButton);

		runningPanel.add(new JLabel(COUNTER_TITLE_STR));
		squashedCounterLabel = new JLabel("0");
		runningPanel.add(squashedCounterLabel);

		myCardPanel.add(START_VIEW, startButtonPanel);
		myCardPanel.add(RUNNING_VIEW, runningPanel);
		myLayout.show(myCardPanel, START_VIEW);

		startButtonSeparatorPanel.add(myCardPanel, BorderLayout.CENTER);
		selectionPanel.add(startButtonSeparatorPanel);

		//setSize(400,300);
		setLocation(100,100);
		pack();
		setVisible(true);
		}
		
	/**
	To show the count of images processed required a jump in complexity including
	input events and threading to allow painting thread to continue showing updates,
	and Runnable methods to be called back from the Swing painting 
	thread to avoid problems with Swing which needs a single thread
	of GUI updates.
	*/
	public void run(){
		// Create a new JPEG directory copier and set it running
		setCursor(new Cursor(Cursor.WAIT_CURSOR));
		myLayout.show(myCardPanel, RUNNING_VIEW);
		//myCardPanel.invalidate();
		//myCardPanel.revalidate();
		boolean desiredDegrain = (Boolean.valueOf(degrainToggle.getSelection().getActionCommand())).booleanValue();
		JPEGDirCopy copier = new JPEGDirCopy(desiredWidth, desiredHeight, desiredQuality, desiredBaseline, desiredUltraQual, desiredDegrain, CopyModel.DEF_THREAD_COUNT_ADJUST);
		copier.addCounterListener(this);
		copier.launchOnAllCores(inputFolder, outputFolder);
		myLayout.show(myCardPanel, START_VIEW);
		setCursor(new Cursor(Cursor.DEFAULT_CURSOR));
		UserComms.alertFrame("Finished", copier.getFinishedMessage());
		}

	void changeInputRequest(){
		JFileChooser chooser = new JFileChooser(lastInputDirName);
		chooser.setFileSelectionMode(JFileChooser.DIRECTORIES_ONLY);
		int result = chooser.showOpenDialog(this);
		if (result == JFileChooser.CANCEL_OPTION){
			return;
			}
		inputFolder = chooser.getSelectedFile();
		lastInputDirName = inputFolder.toString();
		inputDirField.setText(lastInputDirName);
		inputDirField.revalidate();
		this.pack();
		//inputDirField.getParent().validate();
		}

	void changeOutputRequest(){
		if ((outputFolder != null) && (! outputFolder.exists())){
			lastOutputDirName = outputFolder.getParent();
		}
		JFileChooser chooser = new JFileChooser(lastOutputDirName);
		chooser.setFileSelectionMode(JFileChooser.DIRECTORIES_ONLY);
		int result = chooser.showSaveDialog(this);
		if (result == JFileChooser.CANCEL_OPTION){
			return;
			}
		outputFolder = chooser.getSelectedFile();
		lastOutputDirName = outputFolder.toString();
		outputDirField.setText(lastOutputDirName);
		outputDirField.invalidate();
		//outputDirField.getParent().validate();
		this.pack();
		}

	// Check the size arguments
	void changeWidthRequest(){
		int argsWidth;
		try	{
			argsWidth = Integer.parseInt(widthField.getText());
			desiredWidth = argsWidth;
			}
		catch (NumberFormatException numFormEx){
			UserComms.warn("Width range from 20 to 2000");
			}
		}

	void changeHeightRequest(){
		int argsHeight;
		try	{
			argsHeight = Integer.parseInt(heightField.getText());
			desiredHeight = argsHeight;
			}
		catch (NumberFormatException numFormEx){
			UserComms.warn("Height range from 20 to 2000");
			}
		}

	void changeQualityRequest(){
		float argsQuality;
		try	{
			argsQuality = Float.parseFloat(qualityField.getText());
			desiredQuality = argsQuality;
			}
		catch (NumberFormatException numFormEx){
			UserComms.warn("Quality range 0.25 to 0.9");
			}
		}

	public void setCounter(final int newCounter){
		//System.err.println("DEBUG received Adjustment= " + newCount);
		SwingUtilities.invokeLater(new Runnable(){
			public void run (){
				squashedCounterLabel.setText(Integer.toString(newCounter));
				}
			});
		// repaint();
		}

	/**
	 Start the GUI interface here. 
	 It calls the run() method later
	 */
	public static void main(String[] args){

		String usage = "Ignoring unexpected arguments.";

		if (args.length != 0){
			UserComms.warn(usage);
			}
		JPEGCopierGUI gui = new JPEGCopierGUI();
		}

}
	
