package com.jeremyparsons.imaging;

import java.io.File;
import java.io.IOException;
import java.util.List;
import java.util.concurrent.ArrayBlockingQueue;
import java.util.concurrent.CopyOnWriteArrayList;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.RejectedExecutionHandler;
import java.util.concurrent.ThreadPoolExecutor;
import java.util.concurrent.TimeUnit;

/**
This class has been written to automate the production of directories of
graphic images in JPEG format where the images are all scaled to fit
within a particular window size. Scaling in advance helps when communication 
bandwidth is limited or if viewing device has limited memory.

The class can be called via a graphical interface or via the command line 
and the {@link main} method. The directory tree of images to be duplicated 
and the new directory name need to be specified along with the desired dimensions 
and quality. 

Lower quality settings (< 0.7) reduce file sizes and increase image
noise and reduce sharpness. Images with much detail like forests,
mountains and small patterns do not compress well with JPEG.

V2.0 Changed from sun proprietary classes to ImageIO libraries and 
from AffineTransform.getScaleInstance() to Java2D and imgscalr library
V2.1 Changed to multi-threaded JPEG compression
V2.1 Makes a good multi-threaded benchmark if started from a terminal via time !

@see org.imgscalr.Scalr;
@see com.jeremyparsons.imaging.JPEGCopierGUI;
@see https://today.java.net/pub/a/today/2007/04/03/perils-of-image-getscaledinstance.html

Copyright Jeremy D. Parsons 2002,2003,2013,2015
@author Jeremy D. Parsons code@jeremyparsons.com
@version 2.1 Date: 20150110
*/

public class JPEGDirCopy {

	public static String FINISHED_SIZE_MESSAGE = "Total written size= ";
	public static String FINISHED_COUNT_MESSAGE = "Total written count= ";
	
	/* Need a CopyModel to hold configuration and progress information */
	CopyModel copyModel = new CopyModel();
	private final List<CounterListener> listeners = new CopyOnWriteArrayList<CounterListener>();

	public static void main(String[] args){

		String usage = "JPEGDirCopy original_directory_name copied_directory_name width_in_pixels height_in_pixels quality(0.1 - 0.9, "
				+ CopyModel.DEF_QUALITY + " default) maximise_compatibility(true|false) UltraQuality(true|false) degrain(true|false) thread_count_adjust(-1,0,1)";

		// Check the number of arguments
		if (args.length != 9){
			System.err.println("Incorrect number of arguments: " + args.length + "\n" + usage);
			System.exit(-1);
			}
		// Check the directory arguments
		File topDir = new File(args[0]);
		File outDir = new File(args[1]);
		try	{
			if (topDir.getCanonicalPath().equals(outDir.getCanonicalPath())){
				System.err.println("Output directory cannot be the same as input directory.");
				System.exit(-1);
				}
			}
		catch (IOException ioEx){
			System.err.println("Unable to compare the input and output directories to check that they are not the same.\n" + ioEx);
			System.exit(-1);
			}

		// Check the size arguments
		int argsWidth = CopyModel.DEF_OUT_WIDTH;
		int argsHeight = CopyModel.DEF_OUT_HEIGHT;
		float argsQuality = CopyModel.DEF_QUALITY;
		try	{
			argsWidth = Integer.parseInt(args[2]);
			argsHeight = Integer.parseInt(args[3]);
			argsQuality = Float.parseFloat(args[4]);
			}
		catch (NumberFormatException numFormEx){
			System.err.println("Unable to parse one of the input numbers.\n" + numFormEx + "\n" + usage);
			System.exit(-1);
			}
		boolean argsBaseline = Boolean.valueOf(args[5]).booleanValue();
		boolean argsDegrain = Boolean.valueOf(args[6]).booleanValue();
		boolean argsUltraQual = Boolean.valueOf(args[7]).booleanValue();
		int threadCountAdjust = Integer.valueOf(args[8]).intValue();
		

		// Create a new JPEG directory copier and set it running with an ExecutorService
		JPEGDirCopy copier = new JPEGDirCopy(argsWidth, argsHeight, argsQuality, argsBaseline, argsUltraQual, argsDegrain, threadCountAdjust);
		copier.launchOnAllCores(topDir, outDir);
	}
	
	/** Concurrent image compression threads are initialised here */
	public void launchOnAllCores(File topDir, File outDir){
		int coreCount = Runtime.getRuntime().availableProcessors();
		// int maxThreadCount = coreCount + 2 ; // Allow for IO blocking, Should be configurable: affects both memory and speed
		int maxThreadCount = coreCount + copyModel.getDesiredThreadCountAdjust() ; // Allow for IO blocking, Should be configurable: affects both memory and speed
		ArrayBlockingQueue<Runnable> inputTask = new ArrayBlockingQueue<Runnable>(2); // Reduce wasted memory queue=2
		ThreadPoolExecutor taskExecutor = new ThreadPoolExecutor(maxThreadCount, maxThreadCount, 0L, TimeUnit.MILLISECONDS, inputTask);
		//ExecutorService taskExecutor = Executors.newFixedThreadPool(coreCount);
		taskExecutor.setRejectedExecutionHandler(new RejectedExecutionHandler() {
			   public void rejectedExecution(Runnable r, ThreadPoolExecutor executor) {
			      // this will block if the queue is full as opposed to throwing
			      try{
			    	  executor.getQueue().put(r);
			      		} 
			      catch (InterruptedException e) {
					// Ignore ! e.printStackTrace();
				}
			   }
			});
		this.shrinkTree(topDir, outDir, taskExecutor);
		taskExecutor.shutdown();
		try {
			taskExecutor.awaitTermination(Long.MAX_VALUE, TimeUnit.NANOSECONDS);
			}
		catch (InterruptedException intEx) {
			// ignore throw new RuntimeException(intEx);
			}
		//announceWrittenCount(copyModel.getWrittenCount());
		UserComms.inform(getFinishedMessage());
		}

	public JPEGDirCopy(){
		this(CopyModel.DEF_OUT_WIDTH, CopyModel.DEF_OUT_HEIGHT, CopyModel.DEF_QUALITY, CopyModel.DEF_BASELINE_ENCODING, CopyModel.DEF_ULTRA_QUAL, CopyModel.DEF_DEGRAIN, CopyModel.DEF_THREAD_COUNT_ADJUST);
		}

	public JPEGDirCopy(int width, int height, float quality, boolean useBaselineEncoding, boolean doUltraQuality, boolean doBlurring, int threadAdjust){
		this.copyModel.setDesiredWidth(width);
		this.copyModel.setDesiredHeight(height);
		this.copyModel.setDesiredQuality(quality);
		this.copyModel.setDesiredBaseline(useBaselineEncoding);
		this.copyModel.setDesiredUltraQual(doUltraQuality);
		this.copyModel.setDesiredDegrain(doBlurring);
		this.copyModel.setDesiredThreadCountAdjust(threadAdjust);

 		copyModel.setWrittenCount(0);
		}
	/**
	Runs through the directory tree recursively copying and resizing as it goes.
	Will not overwrite any existing files or directories.
	*/
	public void shrinkTree(File top, File topCopy, ExecutorService launcher ){
		// Do not overwrite anything
		if (topCopy.exists()){
			UserComms.warn(topCopy + " already exists !");
			return;
			}
		if (top.isDirectory()){
			if (! topCopy.mkdir()){
				UserComms.warn("Unable to create directory " + topCopy);
				return;
				}
			else	{
				// Get a list of all the files and filter each
				String[] filenamesList = top.list();
				// Go depth first for simplicity of recursion
				for (int i = filenamesList.length - 1; i >= 0; i--){
					this.shrinkTree(new File(top, filenamesList[i]), new File(topCopy, filenamesList[i]), launcher);
					}
				}
			}
		else{
			Runnable copyTask = new ResizeRunnable(top, topCopy, copyModel);
			launcher.execute(copyTask);
			announceWrittenCount(copyModel.getWrittenCount());
			return;
			}
		}
	
	void announceWrittenCount(int count){
		for (CounterListener listener : listeners) {
			listener.setCounter(count);
			}
		}
	
	String getFinishedMessage(){
		String finishedAlertStr = FINISHED_SIZE_MESSAGE 
				+ Long.toString((copyModel.getWrittenSize()) / 1024)
				+ " KB "
				+ FINISHED_COUNT_MESSAGE 
				+ copyModel.getWrittenCount()
				;
		return(finishedAlertStr);
	}

	public void addCounterListener(CounterListener listener){
		listeners.add(listener);
		}
	
	public CopyModel getModel(){
		return(copyModel);
	}

}
	
	
